(* $Id: pax_cmd.ml 771 2008-05-23 02:31:19Z gerd $ *)

open Printf

(* CHECK:

   I have the impression that the test for matching directories works 
   differently in other pax versions. When an archive has a directory
   entry for the directory, matching this entry enables that all
   containing files are matched, too. If there is no directory entry
   for the directory, nothing is matched. I.e. matching is stateful.

   Case 1:

   dir
   dir/file

   case 2:

   dir/file

   _This_ algorithm always uses prefix matching, so for both cases we
   extract the whole directory.

   The manual pages do not explain this.
 *)



let rec test_components re_l path_l pflag =
  match re_l with
    | re :: re_l' ->
	( match path_l with
	    | name :: path_l' ->
		Str.string_match re name 0 &&
		  test_components re_l' path_l' pflag
	    | [] ->
		false
	)
    | [] ->
	pflag || path_l=[]


let slash_re = Str.regexp "/+"

let any_re = Str.regexp ".*"

let compile_pattern nodescent pat_text =
  let pat_split = Str.split_delim slash_re pat_text in
  let pat_split_rev = List.rev pat_split in
  let pat_split', pat_ends_with_slash =    (* / at end is special! *)
    if nodescent then
      (pat_split, false)
    else
      match pat_split_rev with
	| "" :: l -> 
	    (List.rev l, true)
	| l ->
	    (pat_split, false) in
  let pat_parsed =
    List.map
      (fun s ->
	 try 
	   Godi_match.parse_glob_expr ~enable_braces:false s
	 with
	   | _ -> 
	       Godi_match.literal_glob_expr s
      )
      pat_split' in
  let re_list =
    List.map
      (fun g -> Str.regexp(Godi_match.regexp_of_glob_expr g)) 
      pat_parsed in
  let re_list' =
    if pat_ends_with_slash then
      re_list @ [any_re]
    else
      re_list in
  (fun filename ->
     let path = Str.split_delim slash_re filename in
     test_components re_list' path (not nodescent)
  )


let six_months =
  15552000L  (* approximation: 180 days *)


let ls_l now name m =
  (* Produce "ls -l" compatible output *)
  let bits s n =
    sprintf "%c%c%c"
      (if n land 4 <> 0 then 'r' else '-')
      (if n land 2 <> 0 then 'w' else '-')
      (let x =  n land 1 <> 0 in
       match (s,x) with
	 | (false,false) -> '-'
	 | (false,true) -> 'x'
	 | (true,false) -> 'S'
	 | (true,true) -> 's'
      ) in
  let mode = m#mode in
  let mode_str = 
    sprintf "%s%s%s%s"
      (match m#filetype with
	 | `Regular | `Hardlink _ -> "-"
	 | `Directory -> "d"
	 | `Symlink _ -> "l"
	 | `Block _ -> "b"
	 | `Char _ -> "c"
	 | `Fifo -> "p"
	 | _ -> "?"
      )
      (bits (mode land 0o4000 > 0) ((mode land 0o700) lsr 6))
      (bits (mode land 0o2000 > 0) ((mode land 0o070) lsr 3))
      (bits false                   (mode land 0o007))
  in
  let mtime = m#mtime in
  sprintf "%s 1 %s %s %10Ld %s %s%s"
    mode_str
    ( match m#owner_uname with
	| Some s -> s
	| None ->
	    match m#owner_uid with
	      | Some id -> string_of_int id
	      | None -> "0"
    )
    ( match m#owner_gname with
	| Some s -> s
	| None ->
	    match m#owner_gid with
	      | Some id -> string_of_int id
	      | None -> "0"
    )
    m#size
    ( if mtime > now || Int64.sub now mtime >= six_months then
	Netdate.format "%b %e  %Y" (Netdate.create (Int64.to_float mtime))
      else
	Netdate.format "%b %e %H:%M" (Netdate.create (Int64.to_float mtime))
    )
    name
    ( match m#filetype with
	| `Symlink n ->
	    " -> " ^ n
	| `Hardlink n ->
	    " == " ^ n
	| _ ->
	    ""
    )


let main() =
  let now = Int64.of_float(Unix.time()) in

  let read_flag = ref false in
  let write_flag = ref false in
  let complement_flag = ref false in
  let nodescent_flag = ref false in
  let file = ref None in
  let noreplace_flag = ref false in
  let opts = ref "" in
  let privs = ref "" in
  let subst_args = ref [] in
  let update_flag = ref false in
  let verbose_flag = ref false in
  let format = ref None in
  let compression = ref `None in  (* `None implies autodetect *)
  let args = ref [] in
  
  Godi_getopt.parse
    [ "-r", `Set read_flag,
      "Extract archives";

      "-w", `Set write_flag,
      "Create archives";

      "-a", `Unit(fun () -> raise(Arg.Bad ("Unsupported: -a"))),
      "(append - unsupported)";

      "-b", `String(fun _ -> ()),
      "<size>  (Set blocksize - ignored)";

      "-c", `Set complement_flag,
      "Complement file selection";

      "-d", `Set nodescent_flag,
      "Do not descent into directories";

      "-f", `String(fun s -> file := Some s),
      "<file>  Set archive file";

      "-i", `Unit(fun () -> raise(Arg.Bad ("Unsupported: -i"))),
      "(interact - unsupported)";
      
      "-j", `Unit(fun () -> compression := `Bzip2),
      "Process bzip2-compressed archives";

      "-k", `Set noreplace_flag,
      "Do not overwrite files";

      "-l", `Unit(fun () -> raise(Arg.Bad ("Unsupported: -l"))),
      "(create links - unsupported)";

      "-n", `Unit(fun () -> raise(Arg.Bad ("Unsupported: -n"))),
      "(Stop on first matching file - unsupported)";

      "-o", `String (fun s -> opts := s),
      "<opts>  (Set options - none defined)";

      "-p", `String (fun s -> privs := !privs ^ s),
      "<privs>  Privilege flags";

      "-s", `String (fun s -> subst_args := !subst_args @ [s]),
      "<subst>  Do this filename substitution";

      "-t", `Unit(fun () -> raise(Arg.Bad ("Unsupported: -t"))),
      "(leave atime intact - unsupported)";

      "-u", `Set update_flag,
      "Only overwrite with newer files";

      "-v", `Set verbose_flag,
      "Verbose output";

      "-x", `String (fun s -> format := Some s),
      "<fmt>  Set archive format";

      "-X", `Unit(fun () -> raise(Arg.Bad ("Unsupported: -X"))),
      "(limit to single file system - unsupported)";

      "-z", `Unit(fun () -> compression := `Gzip),
      "Process gzip-compressed archives";

      "-h", `Help, "Output help";
      "-?", `Help, "Output help";
      "--", `End_of_switches, "Do not interpret switches beyond this point"
    ]
    (fun arg -> args := !args @ [arg])
    "godi_pax [options] (<file> | <pattern>)...";

  let mode = 
    match !read_flag, !write_flag with
      | false, false -> `List
      | true, false  -> `Extract
      | false, true  -> `Create
      | true, true   -> `Copy in

  let preserve_atime = ref true in
  let preserve_mtime = ref true in
  let preserve_owner = ref false in
  let preserve_filemode = ref false in
  String.iter
    (function
       | 'a' -> preserve_atime := false
       | 'm' -> preserve_mtime := false
       | 'e' -> preserve_atime := true;
	        preserve_mtime := true;
		preserve_owner := true;
		preserve_filemode := true;
       | 'o' -> preserve_owner := true
       | 'p' -> preserve_atime := true;
	        preserve_mtime := true;
		preserve_filemode := true;
       | c ->
	   failwith ("Unknown privilege specified: " ^ String.make 1 c)
    )
    !privs;
  (* Basically, setting privileges is unsupported except for mtime. *)

(*
  if (* !preserve_atime || *) !preserve_owner || !preserve_filemode then
    failwith ("This privilege specification is unsupported: " ^ !privs);
 *)

  let eff_format, compression =
    if mode = `Copy then
      `Tar, `None  (* meaningless *)
    else
      match !format with
	| Some "tar" -> `Tar, !compression
	| Some s -> failwith ("Unknown format: " ^ s)
	| None ->
	    ( (* Look at file name if present: *)
	      match !file with
		| Some name ->
		    let (fmt, compr) =
		      if Filename.check_suffix name ".tar" then
			`Tar, `None
		      else if Filename.check_suffix name ".tar.gz" then
			`Tar, `Gzip
		      else if Filename.check_suffix name ".tgz" then
			`Tar, `Gzip
		      else if Filename.check_suffix name ".tar.bz2" then
			`Tar, `Bzip2
		      else if Filename.check_suffix name ".tbz" then
			`Tar, `Bzip2
		      else if Filename.check_suffix name ".tbz2" then
			`Tar, `Bzip2
		      else
			failwith("Cannot determine format. Specify -x")
		    in
		    (fmt, 
		     if !compression = `None then compr else !compression
		    )
		| None -> failwith("Cannot determine format. Specify -x")
	    )
  in

  let overwrite =
    if !noreplace_flag then
      `Ignore
    else
      if !update_flag then
	`Update
      else
	`Always in
  
  let aargs = Array.of_list !args in
  let have_error = ref false in
  
  ( try
      let subst_list =
	List.map
	  (fun s ->
	     let (lhs,rhs,flags) = Godi_match.split_ed_subst_expr s in
	     let subst_action = Godi_match.parse_ed_subst_expr (lhs,rhs) in
	     let globally = String.contains flags 'g' in
	     let print = String.contains flags 'p' in
	     (fun u ->
		let (success, u') = 
		  Godi_match.substitute2 ~globally subst_action u in
		if success then (
		  if print then prerr_endline (u ^ " >> " ^ u');
		  Some u'
		)
		else
		  None
	     )
	  )
	  !subst_args in
      let rec subst l u =
	match l with
	  | [] -> u
	  | f :: l' ->
	      match f u with
		| None -> subst l' u
		| Some u' -> u'
      in 

      match mode with
	| `List
	| `Extract ->
	    let patterns = 
	      Array.map (compile_pattern !nodescent_flag) aargs in
	    let pat_used =
	      Array.map (fun _ -> false) patterns in
	    let pat_match m =
	      patterns = [| |]  ||  (
		let f = ref false in
		for k = 0 to Array.length patterns - 1 do
		  let p = patterns.(k) m#name in
		  pat_used.(k) <- pat_used.(k) || p;
		  f := !f || p
		done;
		if !complement_flag then not !f else !f
	      )
	    in
	    let in_file =
	      match !file with
		| Some n -> `File n 
		| None -> `Stdin in
	    let ar = Pax.read in_file compression eff_format in
	    let cont = ref true in
	    while !cont do
	      match Pax.read_element ar with
		| None -> cont := false
		| Some el ->
		    let m = Pax.get_metadata el in
		    if pat_match m then (
		      match mode with
			| `List ->
			    let name = subst subst_list m#name in
			    printf "%s\n%!" 
			      ( if !verbose_flag then
				  ls_l now name m
				else
				  name
			      )
			| `Extract ->
			    Pax.extract_element_as_file
			      ~subst:(fun s ->
					let s' = subst subst_list s in
					if !verbose_flag then
					  prerr_endline s';
					s'
				     )
			      ~preserve_mtime:!preserve_mtime
			      ~overwrite
			      el
			| _ ->
			    assert false
		    )
	    done;
	    Pax.postprocess_extraction ar;
	    Pax.close_in ar;
	    for k = 0 to Array.length aargs - 1 do
	      if not pat_used.(k) then (
		prerr_endline("Pattern not matched by any archive member: " ^ 
				aargs.(k));
		have_error := true
	      )
	    done
	      
	| `Create ->
	    let iter =
	      if !args = [] then
		Pax.file_iterator_of_channel
		  ~nodescent:!nodescent_flag
		  (new Netchannels.input_channel stdin)
	      else
		Pax.file_iterator_of_list
		  ~nodescent:!nodescent_flag
		  !args in
	    let out_file =
	      match !file with
		| Some n -> `File n 
		| None -> `Stdout in
	    let ar = Pax.create out_file compression eff_format in

	    let cont = ref true in
	    while !cont do
	      match Pax.next_file iter with
		| Some filename ->
		    Pax.add_element_from_file
		      ~subst:(fun s ->
				let s' = subst subst_list s in
				if !verbose_flag then
				  prerr_endline s';
				s'
			     )
		      ar
		      filename

		| None ->
		    cont := false
	    done;
	    Pax.close_out ar
	      
	| `Copy ->
	    let file_args, dir_arg =
	      match List.rev !args with
		| [] ->
		    failwith "Need a directory argument in copy mode"
		| dir_arg :: args' ->
		    (List.rev args', dir_arg) in
	    let iter =
	      if file_args = [] then
		Pax.file_iterator_of_channel
		  ~nodescent:!nodescent_flag
		  (new Netchannels.input_channel stdin)
	      else
		Pax.file_iterator_of_list
		  ~nodescent:!nodescent_flag
		  file_args in
	    let cp = Pax.create_copier() in

	    let cont = ref true in
	    while !cont do
	      match Pax.next_file iter with
		| Some filename ->
		    Pax.copy_file
		      ~subst:(fun s ->
				let s' = subst subst_list s in
				if !verbose_flag then
				  prerr_endline s';
				s'
			     )
		      ~preserve_mtime:!preserve_mtime
		      ~overwrite
		      cp
		      filename
		      dir_arg

		| None ->
		    cont := false
	    done;
	    Pax.postprocess_copying cp
    with
      | Failure msg ->
	  prerr_endline msg;
	  have_error := true
  );
  if !have_error then exit 1 else exit 0
;;

  
